@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2011-2012 Canonical
+ * Copyright (C) 2011-2013 Canonical
  * Hugely modified parts from powertop-1.13, Copyright 2007, Intel Corporation
  *
  * This program is free software; you can redistribute it and/or
@@ -26,22 +26,36 @@
 #include <unistd.h>
 #include <time.h>
 #include <sys/time.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <limits.h>
+#include <errno.h>
+#include <fcntl.h>
+#include <libgen.h>
+#include <math.h>
 
-#define APP_NAME	"eventstat"
-#define TIMER_STATS	"/proc/timer_stats"
-#define TABLE_SIZE	(251)		/* Should be a prime */
+#define APP_NAME		"eventstat"
+#define TIMER_STATS		"/proc/timer_stats"
+#define TABLE_SIZE		(1997)		/* Should be a prime */
 
-#define OPT_QUIET	(0x00000001)
+#define OPT_QUIET		(0x00000001)
+#define OPT_CUMULATIVE		(0x00000002)
+#define OPT_CMD_SHORT		(0x00000004)
+#define OPT_CMD_LONG		(0x00000008)
+#define OPT_DIRNAME_STRIP	(0x00000010)
+#define OPT_SAMPLE_COUNT	(0x00000020)
+#define OPT_RESULT_STATS	(0x00000040)
+#define OPT_CMD			(OPT_CMD_SHORT | OPT_CMD_LONG)
 
 typedef struct link {
-	void *data;
-	struct link *next;
+	void *data;			/* Data in list */
+	struct link *next;		/* Next item in list */
 } link_t;
 
 typedef struct {
-	link_t	*head;
-	link_t	*tail;
-	size_t	length;
+	link_t	*head;			/* Head of list */
+	link_t	*tail;			/* Tail of list */
+	size_t	length;			/* Length of list */
 } list_t;
 
 typedef void (*list_link_free_t)(void *);
@@ -49,9 +63,11 @@ typedef void (*list_link_free_t)(void *);
 typedef struct timer_info {
 	pid_t		pid;
 	char 		*task;		/* Name of process/kernel task */
+	char		*cmdline;	/* From /proc/$pid/cmdline */
 	char		*func;		/* Kernel waiting func */
 	char		*callback;	/* Kernel timer callback func */
 	char		*ident;		/* Unique identity */
+	bool		kernel_thread;	/* True if task is a kernel thread */
 	unsigned long	total;		/* Total number of events */
 } timer_info_t;
 
@@ -76,12 +92,48 @@ typedef struct sample_delta_list {
 	list_t			list;
 } sample_delta_list_t;
 
-static list_t timer_info_list;			/* cache list of timer_info */
-static list_t sample_list;			/* list of samples, sorted in sample time order */
-static char *csv_results;			/* results in comma separated values */
+static list_t timer_info_list;		/* cache list of timer_info */
+static list_t sample_list;		/* list of samples, sorted in sample time order */
+static char *csv_results;		/* results in comma separated values */
 static volatile bool stop_eventstat = false;	/* set by sighandler */
-static unsigned long opt_threshold;		/* ignore samples with event delta less than this */
-static unsigned int opt_flags;			/* option flags */
+static unsigned long opt_threshold;	/* ignore samples with event delta less than this */
+static unsigned int opt_flags;		/* option flags */
+
+/*
+ *  set_timer_stat()
+ *	enable/disable timer stat
+ */
+void set_timer_stat(const char *str, const bool carp)
+{
+	FILE *fp;
+
+	if ((fp = fopen(TIMER_STATS, "w")) == NULL) {
+		if (carp) {
+			fprintf(stderr, "Cannot write to %s\n",TIMER_STATS);
+			exit(EXIT_FAILURE);
+		} else {
+			return;
+		}
+	}
+	fprintf(fp, "%s\n", str);
+	fclose(fp);
+}
+
+/*
+ *  Stop gcc complaining about no return func
+ */
+void eventstat_exit(const int status) __attribute__ ((noreturn));
+
+/*
+ *  eventstat_exit()
+ *	exit and set timer stat to 0
+ */
+void eventstat_exit(const int status)
+{
+	set_timer_stat("0", false);
+
+	exit(status);
+}
 
 /*
  *  list_init()
@@ -102,19 +154,18 @@ static link_t *list_append(list_t *list, void *data)
 {
 	link_t *link;
 
-	if ((link = calloc(sizeof(link_t), 1)) == NULL) {
+	if ((link = calloc(1, sizeof(link_t))) == NULL) {
 		fprintf(stderr, "Cannot allocate list link\n");
-		exit(EXIT_FAILURE);
+		eventstat_exit(EXIT_FAILURE);
 	}
 	link->data = data;
 
 	if (list->head == NULL) {
 		list->head = link;
-		list->tail = link;
 	} else {
 		list->tail->next = link;
-		list->tail = link;
 	}
+	list->tail = link;
 	list->length++;
 
 	return link;
@@ -124,7 +175,7 @@ static link_t *list_append(list_t *list, void *data)
  *  list_free()
  *	free the list
  */
-static void list_free(list_t *list, list_link_free_t freefunc)
+static void list_free(list_t *list, const list_link_free_t freefunc)
 {
 	link_t	*link, *next;
 
@@ -145,6 +196,8 @@ static void list_free(list_t *list, list_link_free_t freefunc)
  */
 static void handle_sigint(int dummy)
 {
+	(void)dummy;	/* Stop unused parameter warning with -Wextra */
+
 	stop_eventstat = true;
 }
 
@@ -173,7 +226,7 @@ static void samples_free(void)
  *  sample_add()
  *	add a timer_stat's delta and info field to a list at time position whence
  */
-static void sample_add(timer_stat_t *timer_stat, unsigned long whence)
+static void sample_add(timer_stat_t *timer_stat, const unsigned long whence)
 {
 	link_t	*link;
 	bool	found = false;
@@ -198,7 +251,7 @@ static void sample_add(timer_stat_t *timer_stat, unsigned long whence)
 	if (!found) {
 		if ((sdl = calloc(1, sizeof(sample_delta_list_t))) == NULL) {
 			fprintf(stderr, "Cannot allocate sample delta list\n");
-			exit(EXIT_FAILURE);
+			eventstat_exit(EXIT_FAILURE);
 		}
 		sdl->whence = whence;
 		list_append(&sample_list, sdl);
@@ -207,7 +260,7 @@ static void sample_add(timer_stat_t *timer_stat, unsigned long whence)
 	/* Now append the sdi onto the list */
 	if ((sdi = calloc(1, sizeof(sample_delta_item_t))) == NULL) {
 		fprintf(stderr, "Cannot allocate sample delta item\n");
-		exit(EXIT_FAILURE);
+		eventstat_exit(EXIT_FAILURE);
 	}
 	sdi->delta = timer_stat->delta;
 	sdi->info  = timer_stat->info;
@@ -219,7 +272,7 @@ static void sample_add(timer_stat_t *timer_stat, unsigned long whence)
  *  sample_find()
  *	scan through a sample_delta_list for timer info, return NULL if not found
  */
-static sample_delta_item_t inline *sample_find(sample_delta_list_t *sdl, timer_info_t *info)
+inline static sample_delta_item_t *sample_find(sample_delta_list_t *sdl, const timer_info_t *info)
 {
 	link_t *link;
 
@@ -243,6 +296,72 @@ static int info_compare_total(const void *item1, const void *item2)
 	return (*info2)->total - (*info1)->total;
 }
 
+/*
+ *  pid_a_kernel_thread
+ *
+ */
+static bool pid_a_kernel_thread(const pid_t id)
+{
+	char buffer[128];
+	char path[32];
+
+	snprintf(buffer, sizeof(buffer), "/proc/%d/exe", id);
+	if (readlink(buffer, path, sizeof(path)) < 0)
+		if (errno == ENOENT)
+			return true;
+
+	return false;
+}
+
+/*
+ *  get_pid_cmdline
+ * 	get process's /proc/pid/cmdline
+ */
+static char *get_pid_cmdline(const pid_t id)
+{
+	char buffer[4096];
+	char *ptr;
+	int fd;
+	ssize_t ret;
+
+	snprintf(buffer, sizeof(buffer), "/proc/%d/cmdline", id);
+
+	if ((fd = open(buffer, O_RDONLY)) < 0)
+		return NULL;
+
+	if ((ret = read(fd, buffer, sizeof(buffer))) <= 0)
+		return NULL;
+
+	close(fd);
+
+	buffer[sizeof(buffer)-1] = '\0';
+
+	/*
+	 *  OPT_CMD_LONG option we get the full cmdline args
+	 */
+	if (opt_flags & OPT_CMD_LONG) {
+		for (ptr = buffer; ptr < buffer + ret - 1; ptr++) {
+			if (*ptr == '\0')
+				*ptr = ' ';
+		}
+		*ptr = '\0';
+	}
+	/*
+	 *  OPT_CMD_SHORT option we discard anything after a space
+	 */
+	if (opt_flags & OPT_CMD_SHORT) {
+		for (ptr = buffer; *ptr && (ptr < buffer + ret); ptr++) {
+			if (*ptr == ' ')
+				*ptr = '\0';
+		}
+	}
+
+	if (opt_flags & OPT_DIRNAME_STRIP)
+		return strdup(basename(buffer));
+
+	return strdup(buffer);
+}
+
 /*
  *  samples_dump()
  *	dump out collected sample information
@@ -252,9 +371,11 @@ static void samples_dump(const char *filename, const int duration)
 	sample_delta_list_t	*sdl;
 	timer_info_t **sorted_timer_infos;
 	link_t	*link;
-	int i = 0;
+	size_t i = 0;
 	size_t n = timer_info_list.length;
 	FILE *fp;
+	unsigned long count = 0;
+	double dur;
 
 	if (filename == NULL)
 		return;
@@ -266,7 +387,7 @@ static void samples_dump(const char *filename, const int duration)
 
 	if ((sorted_timer_infos = calloc(n, sizeof(timer_info_t*))) == NULL) {
 		fprintf(stderr, "Cannot allocate buffer for sorting timer_infos\n");
-		exit(EXIT_FAILURE);
+		eventstat_exit(EXIT_FAILURE);
 	}
 
 	/* Just want the timers with some non-zero total */
@@ -279,36 +400,112 @@ static void samples_dump(const char *filename, const int duration)
 	qsort(sorted_timer_infos, n, sizeof(timer_info_t *), info_compare_total);
 
 	fprintf(fp, "Task:");
-	for (i=0; i<n; i++)
-		fprintf(fp, ",%s", sorted_timer_infos[i]->task);
+	for (i = 0; i < n; i++) {
+		char *task;
+
+		if ((opt_flags & OPT_CMD) && (sorted_timer_infos[i]->cmdline != NULL))
+			task = sorted_timer_infos[i]->cmdline;
+		else
+			task = sorted_timer_infos[i]->task;
+
+		fprintf(fp, ",%s", task);
+	}
 	fprintf(fp, "\n");
 
 	fprintf(fp, "Init Function:");
-	for (i=0; i<n; i++)
+	for (i = 0; i < n; i++)
 		fprintf(fp, ",%s", sorted_timer_infos[i]->func);
 	fprintf(fp, "\n");
 
 	fprintf(fp, "Callback:");
-	for (i=0; i<n; i++)
+	for (i = 0; i < n; i++)
 		fprintf(fp, ",%s", sorted_timer_infos[i]->callback);
 	fprintf(fp, "\n");
 
 	fprintf(fp, "Total:");
-	for (i=0; i<n; i++)
+	for (i = 0; i < n; i++)
 		fprintf(fp, ",%lu", sorted_timer_infos[i]->total);
 	fprintf(fp, "\n");
 
+	/*
+	 *  duration - if -C option is used then don't scale by the
+	 *  per sample duration time, instead give the raw sample count
+	 *  by scaling by 1.0 (i.e. no scaling).
+	 */
+	dur = (opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : (double)duration;
+
 	for (link = sample_list.head; link; link = link->next) {
+		count++;
 		sdl = (sample_delta_list_t*)link->data;
 		fprintf(fp, "%lu", sdl->whence);
 
 		/* Scan in timer info order to be consistent for all sdl rows */
-		for (i=0; i<n; i++) {
+		for (i = 0; i < n; i++) {
 			sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
 			if (sdi)
-				fprintf(fp,",%f", (double)sdi->delta / (double)duration);
+				fprintf(fp, ",%f", (double)sdi->delta / dur);
 			else
-				fprintf(fp,",");
+				fprintf(fp, ",%f", 0.0);
+		}
+		fprintf(fp, "\n");
+	}
+
+	/*
+	 *  -S option - some statistics, min, max, average, std.dev.
+	 */
+	if (opt_flags & OPT_RESULT_STATS) {
+		fprintf(fp, "Min:");
+		for (i = 0; i < n; i++) {
+			unsigned long min = ~0;
+
+			for (link = sample_list.head; link; link = link->next) {
+				sdl = (sample_delta_list_t*)link->data;
+				sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
+				if (sdi && min > sdi->delta)
+					min = sdi->delta;
+			}
+			fprintf(fp, ",%f", (double)min / dur);
+		}
+		fprintf(fp, "\n");
+
+		fprintf(fp, "Max:");
+		for (i = 0; i < n; i++) {
+			unsigned long max = 0;
+
+			for (link = sample_list.head; link; link = link->next) {
+				sdl = (sample_delta_list_t*)link->data;
+				sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
+				if (sdi && max < sdi->delta)
+					max = sdi->delta;
+			}
+			fprintf(fp, ",%f", (double)max / dur);
+		}
+		fprintf(fp, "\n");
+
+		fprintf(fp, "Average:");
+		for (i = 0; i < n; i++)
+			fprintf(fp, ",%f", ((double)sorted_timer_infos[i]->total / dur) / (double)count);
+		fprintf(fp, "\n");
+
+		/*
+		 *  population standard deviation
+		 */
+		fprintf(fp, "Std.Dev.:");
+		for (i = 0; i < n; i++) {
+			double average = (double)sorted_timer_infos[i]->total / (double)count;
+			double sum = 0.0;
+
+			for (link = sample_list.head; link; link = link->next) {
+				sdl = (sample_delta_list_t*)link->data;
+				sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
+				if (sdi) {
+					double diff = ((double)sdi->delta - average) / dur;
+					diff = diff * diff;
+					sum += diff;
+				}
+			}
+			sum = sum / (double)count;
+			fprintf(fp, ",%f", sqrt(sum));
 		}
 		fprintf(fp, "\n");
 	}
@@ -322,7 +519,7 @@ static void samples_dump(const char *filename, const int duration)
  *	try to find existing timer info in cache, and to the cache
  *	if it is new.
  */
-static timer_info_t *timer_info_find(timer_info_t *new_info)
+static timer_info_t *timer_info_find(const timer_info_t *new_info)
 {
 	link_t *link;
 	timer_info_t *info;
@@ -335,21 +532,26 @@ static timer_info_t *timer_info_find(timer_info_t *new_info)
 
 	if ((info = calloc(1, sizeof(timer_info_t))) == NULL) {
 		fprintf(stderr, "Cannot allocate timer info\n");
-		exit(EXIT_FAILURE);
+		eventstat_exit(EXIT_FAILURE);
 	}
 
 	info->pid = new_info->pid;
 	info->task = strdup(new_info->task);
+	if (opt_flags & OPT_CMD)
+		info->cmdline = get_pid_cmdline(new_info->pid);
+
 	info->func = strdup(new_info->func);
 	info->callback = strdup(new_info->callback);
 	info->ident = strdup(new_info->ident);
+	info->kernel_thread = new_info->kernel_thread;
+	info->total = 0;
 
 	if (info->task == NULL ||
 	    info->func == NULL ||
 	    info->callback == NULL ||
 	    info->ident == NULL) {
 		fprintf(stderr, "Out of memory allocating a timer stat fields\n");
-		exit(1);
+		eventstat_exit(EXIT_FAILURE);
 	}
 
 	/* Does not exist in list, append it */
@@ -368,6 +570,7 @@ static void timer_info_free(void *data)
 	timer_info_t *info = (timer_info_t*)data;
 
 	free(info->task);
+	free(info->cmdline);
 	free(info->func);
 	free(info->callback);
 	free(info->ident);
@@ -387,7 +590,7 @@ static void timer_info_list_free(void)
  *  hash_pjw()
  *	Hash a string, from Aho, Sethi, Ullman, Compiling Techniques.
  */
-static unsigned long hash_pjw(char *str)
+static unsigned long hash_pjw(const char *str)
 {
   	unsigned long h=0, g;
 
@@ -412,7 +615,7 @@ static void timer_stat_free_contents(
 {
 	int i;
 
-	for (i=0; i<TABLE_SIZE; i++) {
+	for (i = 0; i < TABLE_SIZE; i++) {
 		timer_stat_t *ts = timer_stats[i];
 
 		while (ts) {
@@ -432,11 +635,12 @@ static void timer_stat_free_contents(
  */
 static void timer_stat_add(
 	timer_stat_t *timer_stats[],	/* timer stat hash table */
-	unsigned long count,		/* event count */
-	pid_t pid,			/* PID of task */
+	const unsigned long count,	/* event count */
+	const pid_t pid,		/* PID of task */
 	char *task,			/* Name of task */
 	char *func,			/* Kernel function */
-	char *callback)			/* Kernel timer callback */
+	char *callback,			/* Kernel timer callback */
+	const bool kernel_thread)	/* Is a kernel thread */
 {
 	char buf[4096];
 	timer_stat_t *ts;
@@ -458,7 +662,7 @@ static void timer_stat_add(
 
 	if ((ts_new = malloc(sizeof(timer_stat_t))) == NULL) {
 		fprintf(stderr, "Out of memory allocating a timer stat\n");
-		exit(1);
+		eventstat_exit(EXIT_FAILURE);
 	}
 
 	info.pid = pid;
@@ -466,6 +670,7 @@ static void timer_stat_add(
 	info.func = func;
 	info.callback = callback;
 	info.ident = buf;
+	info.kernel_thread = kernel_thread;
 
 	ts_new->count  = count;
 	ts_new->info = timer_info_find(&info);
@@ -506,9 +711,16 @@ static void timer_stat_sort_freq_add(
 	timer_stat_t *new)		/* timer stat to add */
 {
 	while (*sorted) {
-		if ((*sorted)->delta < new->delta) {
-			new->sorted_freq_next = *(sorted);
-			break;
+		if (opt_flags & OPT_CUMULATIVE) {
+			if ((*sorted)->count < new->count) {
+				new->sorted_freq_next = *(sorted);
+				break;
+			}
+		} else {
+			if ((*sorted)->delta < new->delta) {
+				new->sorted_freq_next = *(sorted);
+				break;
+			}
 		}
 		sorted = &(*sorted)->sorted_freq_next;
 	}
@@ -524,17 +736,18 @@ static void timer_stat_sort_freq_add(
 static void timer_stat_diff(
 	const int duration,		/* time between each sample */
 	const int n_lines,		/* number of lines to output */
-	unsigned long whence,		/* nth sample */
+	const unsigned long whence,	/* nth sample */
 	timer_stat_t *timer_stats_old[],/* old timer stats samples */
 	timer_stat_t *timer_stats_new[])/* new timer stats samples */
 {
 	int i;
 	int j = 0;
 	unsigned long total = 0UL;
+	unsigned long kt_total = 0UL;
 
 	timer_stat_t *sorted = NULL;
 
-	for (i=0; i<TABLE_SIZE; i++) {
+	for (i = 0; i < TABLE_SIZE; i++) {
 		timer_stat_t *ts;
 
 		for (ts = timer_stats_new[i]; ts; ts = ts->next) {
@@ -560,27 +773,40 @@ static void timer_stat_diff(
 	}
 
 	if (!(opt_flags & OPT_QUIET)) {
-		printf("%1s %6s %-5s %-15s %-25s %-s\n",
-			"", "Evnt/s", "PID", "Task", "Init Function", "Callback");
+		printf("%8s %-5s %-15s %-25s %-s\n",
+			opt_flags & OPT_CUMULATIVE ? "Events" : "Event/s",
+			"PID", "Task", "Init Function", "Callback");
 
 		while (sorted) {
-			if ((n_lines == -1) || (j < n_lines)) {
+			if (((n_lines == -1) || (j < n_lines)) && (sorted->delta != 0)) {
 				j++;
-				printf("%1s %6.2f %5d %-15s %-25s %-s\n",
-					sorted->old ? " " : "N",
-					(double)sorted->delta / (double)duration,
-					sorted->info->pid, sorted->info->task,
-					sorted->info->func, sorted->info->callback);
+				if (opt_flags & OPT_CUMULATIVE) {
+					printf("%1s %6lu %5d %-15s %-25s %-s\n",
+						sorted->old ? " " : "N",
+						sorted->count,
+						sorted->info->pid, sorted->info->task,
+						sorted->info->func, sorted->info->callback);
+				} else {
+					printf("%1s %6.2f %5d %-15s %-25s %-s\n",
+						sorted->old ? " " : "N",
+						(double)sorted->delta / (double)duration,
+						sorted->info->pid, sorted->info->task,
+						sorted->info->func, sorted->info->callback);
+				}
 			}
 			total += sorted->delta;
+			if (sorted->info->kernel_thread)
+				kt_total += sorted->delta;
+
 			sorted = sorted->sorted_freq_next;
 		}
-		printf("%lu Total events, %5.2f events/sec\n\n",
-			total, (double)total / duration);
+		printf("%lu Total events, %5.2f events/sec (kernel: %5.2f, userspace: %5.2f)\n\n",
+			total, (double)total / duration,
+			(double)kt_total / duration,
+			(double)(total - kt_total) / duration);
 	}
 }
 
-
 /*
  *  get_events()
  *	scan /proc/timer_stats and populate a timer stat hash table with
@@ -604,6 +830,7 @@ void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
 		char task[64];
 		char func[128];
 		char timer[128];
+		bool kernel_thread;
 
 		if (fgets(buf, sizeof(buf), fp) == NULL)
 			break;
@@ -629,6 +856,15 @@ void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
 		sscanf(buf, "%lu", &count);
 		sscanf(ptr, "%d %s %s (%[^)])", &pid, task, func, timer);
 
+		kernel_thread = pid_a_kernel_thread(pid);
+
+		if (kernel_thread) {
+			char tmp[64];
+			task[13] = '\0';
+			snprintf(tmp, sizeof(tmp), "[%s]", task);
+			strcpy(task, tmp);
+		}
+
 		if (strcmp(task, "swapper") == 0 &&
 		    strcmp(func, "hrtimer_start_range_ns") == 0 &&
 		    strcmp(timer, "tick_sched_timer") == 0) {
@@ -648,40 +884,31 @@ void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
 		    (strcmp(task, APP_NAME) == 0))
 			continue;
 
-		timer_stat_add(timer_stats, count, pid, task, func, timer);
+		timer_stat_add(timer_stats, count, pid, task, func, timer, kernel_thread);
 	}
 
 	fclose(fp);
 }
 
-/*
- *  set_timer_stat()
- *	enable/disable timer stat
- */
-void set_timer_stat(char *str)
-{
-	FILE *fp;
-
-	if ((fp = fopen(TIMER_STATS, "w")) == NULL) {
-		fprintf(stderr, "Cannot write to %s\n",TIMER_STATS);
-		exit(EXIT_FAILURE);
-	}
-	fprintf(fp, "%s\n", str);
-	fclose(fp);
-}
-
 /*
  *  show_usage()
  *	show how to use
  */
 void show_usage(void)
 {
-	printf("Usage: %s [-q] [-r csv_file] [-n event_count] [duration] [count]\n", APP_NAME);
-	printf("\t-h help\n");
-	printf("\t-n specifies number of events to display\n");
-	printf("\t-q run quietly, useful with option -r\n");
-	printf("\t-r specifies a comma separated values output file to dump samples into.\n");
-	printf("\t-t specifies an event threshold where samples less than this are ignored.\n");
+	printf("%s, version %s\n\n", APP_NAME, VERSION);
+	printf("Usage: %s [options] [duration] [count]\n", APP_NAME);
+	printf("Options are:\n");
+	printf("  -c\t\treport cumulative events rather than events per second.\n");
+	printf("  -C\t\treport event count rather than event per second in CSV output.\n");
+	printf("  -h\t\tprint this help.\n");
+	printf("  -l\t\tuse long cmdline text from /proc/pid/cmdline in CSV output.\n");
+	printf("  -n events\tspecifies number of events to display.\n");
+	printf("  -q\t\trun quietly, useful with option -r.\n");
+	printf("  -r filename\tspecifies a comma separated values (CSV) output file to dump samples into.\n");
+	printf("  -s\t\tuse short process name from /proc/pid/cmdline in CSV output.\n");
+	printf("  -S\t\tcalculate min, max, average and standard deviation in CSV output.\n");
+	printf("  -t threshold\tsamples less than the specified threshold are ignored.\n");
 }
 
 int main(int argc, char **argv)
@@ -698,25 +925,34 @@ int main(int argc, char **argv)
 	list_init(&sample_list);
 
 	for (;;) {
-		int c = getopt(argc, argv, "hn:qr:t:");
+		int c = getopt(argc, argv, "cCdsSlhn:qr:t:");
 		if (c == -1)
 			break;
 		switch (c) {
+		case 'c':
+			opt_flags |= OPT_CUMULATIVE;
+			break;
+		case 'C':
+			opt_flags |= OPT_SAMPLE_COUNT;
+			break;
+		case 'd':
+			opt_flags |= OPT_DIRNAME_STRIP;
+			break;
 		case 'h':
 			show_usage();
-			exit(EXIT_SUCCESS);
+			eventstat_exit(EXIT_SUCCESS);
 		case 'n':
 			n_lines = atoi(optarg);
 			if (n_lines < 1) {
 				fprintf(stderr, "-n option must be greater than 0\n");
-				exit(EXIT_FAILURE);
+				eventstat_exit(EXIT_FAILURE);
 			}
 			break;
 		case 't':
 			opt_threshold = strtoull(optarg, NULL, 10);
 			if (opt_threshold < 1) {
 				fprintf(stderr, "-t threshold must be 1 or more.\n");
-				exit(EXIT_FAILURE);
+				eventstat_exit(EXIT_FAILURE);
 			}
 			break;
 		case 'q':
@@ -725,6 +961,15 @@ int main(int argc, char **argv)
 		case 'r':
 			csv_results = optarg;
 			break;
+		case 's':
+			opt_flags |= OPT_CMD_SHORT;
+			break;
+		case 'S':
+			opt_flags |= OPT_RESULT_STATS;
+			break;
+		case 'l':
+			opt_flags |= OPT_CMD_LONG;
+			break;
 		}
 	}
 
@@ -732,7 +977,7 @@ int main(int argc, char **argv)
 		duration = atoi(argv[optind++]);
 		if (duration < 1) {
 			fprintf(stderr, "Duration must be > 0\n");
-			exit(EXIT_FAILURE);
+			eventstat_exit(EXIT_FAILURE);
 		}
 	}
 
@@ -741,7 +986,7 @@ int main(int argc, char **argv)
 		count = atoi(argv[optind++]);
 		if (count < 1) {
 			fprintf(stderr, "Count must be > 0\n");
-			exit(EXIT_FAILURE);
+			eventstat_exit(EXIT_FAILURE);
 		}
 	}
 
@@ -750,17 +995,23 @@ int main(int argc, char **argv)
 	if (geteuid() != 0) {
 		fprintf(stderr, "%s requires root privileges to write to %s\n",
 			APP_NAME, TIMER_STATS);
-		exit(EXIT_FAILURE);
+		eventstat_exit(EXIT_FAILURE);
 	}
 
 	signal(SIGINT, &handle_sigint);
 
-	/* Should really catch signals and set back to zero before we die */
-	set_timer_stat("1");
-	sleep(1);
+	if ((timer_stats_old = calloc(TABLE_SIZE, sizeof(timer_stat_t*))) == NULL) {
+		fprintf(stderr, "Cannot allocate old timer stats table\n");
+		eventstat_exit(EXIT_FAILURE);
+	}
+	if ((timer_stats_new = calloc(TABLE_SIZE, sizeof(timer_stat_t*))) == NULL) {
+		fprintf(stderr, "Cannot allocate old timer stats table\n");
+		eventstat_exit(EXIT_FAILURE);
+	}
 
-	timer_stats_old = calloc(TABLE_SIZE, sizeof(timer_stat_t*));
-	timer_stats_new = calloc(TABLE_SIZE, sizeof(timer_stat_t*));
+	/* Should really catch signals and set back to zero before we die */
+	set_timer_stat("1", true);
+	sleep(1);
 
 	gettimeofday(&tv1, NULL);
 	get_events(timer_stats_old);
@@ -797,7 +1048,5 @@ int main(int argc, char **argv)
 	samples_free();
 	timer_info_list_free();
 
-	set_timer_stat("0");
-
-	exit(EXIT_SUCCESS);
+	eventstat_exit(EXIT_SUCCESS);
 }
